from fatty_acylizer.models.gradient.model import DiffCounter


def test_counting():
    counter = DiffCounter(list('aaab'))
    assert counter['a'] == 3


def test_factor():
    counter = DiffCounter(list('aaab'))
    assert counter.factor == 1


def test_npermutations():
    counter = DiffCounter(list('aaab'))
    assert counter.n_permutations == 4


def test_differentiation_to_a():
    counter = DiffCounter(list('aaab'))
    expected = DiffCounter(list('aab'))
    expected.factor = 3
    assert counter.differentiate('a') == expected


def test_differentiation_to_b():
    counter = DiffCounter(list('aaab'))
    expected = DiffCounter(list('aaa'))
    assert counter.differentiate('b') == expected


def test_differentiation_to_c():
    counter = DiffCounter(list('aaab'))
    assert counter.differentiate('c') is None



import collections
import itertools as it
import math
import typing
from collections import defaultdict

import memory_profiler
from fatty_acylizer.util.combinatorics import calculate_number_of_permutations
from fatty_acylizer.util.feature import BasicNumber
from fatty_acylizer.util.model import Profile
from fatty_acylizer.util.model import Result


class DiffCounter(collections.Counter):
    def __init__(self, *args, factor: int = 1, n_permutations: int = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.factor = factor
        if not n_permutations:
            n_permutations = calculate_number_of_permutations(self)
        self.n_permutations = n_permutations

    def differentiate(self, with_respect_to: BasicNumber) -> typing.Self:
        copy = self.copy()
        factor = copy[with_respect_to]
        if factor == 0:
            return None
        elif factor == 1:
            copy.pop(with_respect_to)
        else:
            copy[with_respect_to] -= 1
        copy.factor *= factor
        return copy


Space = dict[BasicNumber, list[DiffCounter]]


class Model:
    @memory_profiler.profile()
    def __init__(self, fatty_acids: list[BasicNumber], n_fa_per_lipid: int):
        self.jac = True
        self.n_fa_per_lipid = n_fa_per_lipid
        lipid_space = defaultdict(list)
        for combination in it.combinations_with_replacement(
            fatty_acids, r=n_fa_per_lipid
        ):
            sum_feature = sum(combination)
            lipid_space[sum_feature].append(DiffCounter(combination))
        self.lipid_space: Space = dict(lipid_space)

        self.gradient: list[Space] = []
        for fatty_acid in fatty_acids:
            diff_space = {}
            for lipid, combinations in lipid_space.items():
                diffs = (
                    comb.differentiate(with_respect_to=fatty_acid)
                    for comb in combinations
                )
                diff_space[lipid] = [comb for comb in diffs if comb is not None]
            self.gradient.append(diff_space)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        params = ', '.join([f'{key}={value}' for key, value in self.settings.items()])
        return f'{cls_name}({params})'

    @property
    def settings(self):
        return {
            'jac': self.jac,
            'n_fa_per_lipid': self.n_fa_per_lipid,
            'fatty_acids': self.fatty_acids,
        }

    def compute_profile_gradient(self, fa_profile: Profile) -> Result:
        lipid_space = compute_space(self.lipid_space, fa_profile)
        gradient = [compute_space(g, fa_profile) for g in self.gradient]
        return Result(lipid_space, gradient)

    def __call__(self, fa_profile: Profile) -> Result:
        return self.compute_profile_gradient(fa_profile=fa_profile)


def compute_space(
    lipid_space: Space,
    fa_profile: Profile,
) -> Profile:
    prob_space = {}
    for key, combinations in lipid_space.items():
        probability = 0
        for comb in combinations:
            try:
                prob = math.prod(
                    fa_profile[key] ** times for key, times in comb.items()
                )
            except KeyError:
                # prob = 0
                pass
            else:
                weighted_prob = prob * comb.factor * comb.n_permutations
                probability += weighted_prob
        prob_space[key] = probability
    return prob_space


if __name__ == '__main__':
    import pathlib

    with open(
        pathlib.Path(__file__).parent.parent.parent / 'data/fatty_acids.csv'
    ) as f:
        lines = (line.strip().split(',') for line in f)
        fatty_acids = [complex(real=int(real), imag=int(imag)) for real, imag in lines]

    model = Model(fatty_acids=fatty_acids, n_fa_per_lipid=3)



from copy import copy
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import NamedTuple
from typing import Protocol

import numpy as np
import numpy.typing as npt
import scipy.optimize
from fatty_acylizer.util.feature import BasicNumber
from fatty_acylizer.util.model import Model
from fatty_acylizer.util.model import Profile
from fatty_acylizer.util.model import Result
from sklearn.feature_extraction import DictVectorizer


@dataclass
class Callback:
    x: npt.NDArray
    fun: float
    nfev: int | None = field(default=None)
    message: str | None = field(default=None)
    jac: npt.NDArray | None = field(default=None)
    njev: int | None = field(default=None)
    hess: npt.NDArray | None = field(default=None)
    nhev: int | None = field(default=None)
    nit: int | None = field(default=None)
    success: bool | None = field(default=None)
    status: int | None = field(default=None)

    @classmethod
    def from_optimize_result(cls, result):
        kwargs = {
            k: copy(v) for k, v in result.items() if k in cls.__dataclass_fields__
        }
        return cls(**kwargs)

    def report(self) -> dict[str, str]:
        return {k: v for k, v in self.__dict__.items() if v is not None}


class Serializer(Protocol):
    def __call__(
        self,
        callback: Callback,
    ): ...


class ArrayResult(NamedTuple):
    profile: npt.NDArray
    gradient: list[npt.NDArray] | None


class Optimization:
    def __init__(
        self,
        model: Model,
        ftol: float = 5e-3,
        gtol: float = 1e-4,
        verbose: bool = False,
        max_iter: int = 100,
    ):
        self.raw_model = model
        self.ftol = ftol
        self.verbose = verbose
        self.max_iter = max_iter
        self.gtol = gtol

        self.method: str = 'l-bfgs-b'
        # require setup!
        self._lipid_vectorizer: DictVectorizer | None = None
        self._fa_vectorizer: DictVectorizer | None = None
        self._target_array: npt.NDArray | None = None
        self._x0: npt.NDArray | None = None

        self._history: list[Callback] = []
        self._callback_handlers: list = []

    @property
    def lipid_vectorizer(self) -> DictVectorizer:
        if self._lipid_vectorizer is None:
            raise NotImplementedError
        return self._lipid_vectorizer

    @property
    def fa_vectorizer(self) -> DictVectorizer:
        if self._fa_vectorizer is None:
            raise NotImplementedError
        return self._fa_vectorizer

    @property
    def target_array(self) -> npt.NDArray:
        if self._target_array is None:
            raise NotImplementedError
        return self._target_array

    @property
    def x0(self) -> npt.NDArray:
        if self._x0 is None:
            raise NotImplementedError
        return self._x0

    @property
    def history(self) -> list[Callback]:
        return self._history

    def model(self, fa_array: npt.NDArray) -> ArrayResult:
        """Wrapper around the model

        Utilizes the initialized DictVectorizers to translate profiles to array
        representation.
        """
        fa_profile: Profile = self.fa_vectorizer.inverse_transform(
            fa_array.reshape(1, -1)
        )[0]
        result: Result = self.raw_model(
            fa_profile=fa_profile,
        )
        lipid_array: npt.NDArray = self.lipid_vectorizer.transform(
            result.profile
        ).reshape(-1)
        if not result.gradient:
            return ArrayResult(profile=lipid_array, gradient=None)

        gradient = [
            self.lipid_vectorizer.transform(g).reshape(-1) for g in result.gradient
        ]
        return ArrayResult(profile=lipid_array, gradient=gradient)

    def objective_function(
        self,
        fa_array: npt.NDArray,
    ) -> float | tuple[float, npt.NDArray]:
        result = self.model(fa_array=fa_array)
        lipid_array = result.profile
        distance_squared = np.sum(np.square(lipid_array - self.target_array))
        constraint = (1 - np.sum(fa_array)) ** 2
        function_value: float = float(distance_squared + constraint)
        if not result.gradient:
            return function_value
        gradient = result.gradient
        g_distance_sq = 2 * np.sum((lipid_array - self.target_array) * gradient, axis=1)
        g_constraint = 2 * np.sum(fa_array) - 2
        gradient_value: npt.NDArray = g_distance_sq + g_constraint
        return (function_value, gradient_value)

    def fit(
        self,
        fa_profile: dict[BasicNumber, float],
        target_profile: dict[BasicNumber, float],
        requires_setup: bool = True,
    ):
        """Fit a fatty acid profile to a target lipid profile.

        args:
            fa_profile:
                start guess for the optimizer
            target_profile:
                lipid profile the fatty acid profile is fitted to

        This function has important side effects and initializes the start guess x0,
        the target lipid profile, and the vectorizers for mapping profiles to arrays and
        back.
        """
        if requires_setup:
            self.setup(fa_profile=fa_profile, target_profile=target_profile)

        match self.objective_function(self.x0):
            case (float() as value, _):
                distance = value
            case float() as value:
                distance = value
            case _:
                raise ValueError

        start = scipy.optimize.OptimizeResult(
            x=self.x0,
            fun=distance,
        )
        self.callback(start)
        # fit
        result = scipy.optimize.minimize(
            x0=self.x0,
            fun=self.objective_function,
            jac=self.raw_model.jac,
            bounds=[(0, 1)] * len(self.x0),
            method=self.method,
            options=self.opt_settings,
            callback=self.callback,
        )
        self.history.pop()
        self.history.append(Callback.from_optimize_result(result))
        return result

    def setup(
        self,
        fa_profile: dict[BasicNumber, float],
        target_profile: dict[BasicNumber, float],
    ):
        # setup
        self._fa_vectorizer = DictVectorizer(sparse=False, sort=False)
        self._x0 = self.fa_vectorizer.fit_transform([fa_profile]).reshape(-1)

        # This always constructs all possible lipids, even if their probability is 0
        # This way, the lipid vectorizer is always properly defined
        result = self.raw_model(fa_profile=fa_profile)
        lipid_profile = result.profile
        self._lipid_vectorizer = DictVectorizer(sparse=False, sort=False)
        self.lipid_vectorizer.fit([lipid_profile])
        self._target_array = self.lipid_vectorizer.transform(target_profile).reshape(-1)

    def callback(self, intermediate_result: scipy.optimize.OptimizeResult):
        callback = Callback.from_optimize_result(intermediate_result)
        self.history.append(callback)
        for handler in self._callback_handlers:
            handler(callback)

    @property
    def opt_settings(self):
        return {
            'disp': self.verbose,
            'iprint': 101,
            'ftol': self.ftol,
            'gtol': self.gtol,
            'maxiter': self.max_iter,
        }

    @property
    def settings(self):
        settings: dict[str, Any] = self.opt_settings | self.raw_model.settings
        return settings

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        params = ', '.join(
            [f'{key}={value}' for key, value in self.opt_settings.items()]
        )
        return f'{cls_name}({params})'

    def add_callback_handler(self, handler) -> None:
        self._callback_handlers.append(handler)

    def save_results(self, *serializers: Serializer) -> None:
        """Save the results of the optimization run.

        Either the provided serializers or all serializers added through
        `add_callback_handler` will be used for serializing the callbacks
        stored in `history`.
        """

        for serializer in serializers or self._callback_handlers:
            for callback in self.history:
                serializer(callback)



import itertools
from collections import defaultdict


def split(n: int, parts: int, force_even: bool = True) -> list[int]:
    """Splits up the number n into parts."""
    avg = n // parts
    rem = int(n % parts)
    if not force_even:
        return [avg] * (parts - rem) + [avg + 1] * rem

    if rem:
        sub = int(avg - avg % 2)
        return [
            sub,
            *split(
                n - sub,
                parts - 1,
                force_even=force_even,
            ),
        ]

    half = (parts - rem) // 2
    final_rem = int((parts - rem) % 2)
    return [avg - avg % 2] * half + [avg] * final_rem + [avg + avg % 2] * half


def suggest_start_conditions(
    lipid_profile: dict[complex, float],
    n_fa_per_lipid: int,
    fatty_acids: list[complex],
    prefer_even_chains: bool = True,
) -> dict[complex, float]:
    fa_profile: dict[complex, float] = defaultdict(float)
    for lipid, probability in lipid_profile.items():
        if not lipid.imag.is_integer() or not lipid.real.is_integer():
            raise ValueError(
                f'{lipid} may only contain integer values for real and imaginary parts'
            )
        cc = split(
            n=int(lipid.real), parts=n_fa_per_lipid, force_even=prefer_even_chains
        )
        db = split(n=int(lipid.imag), parts=n_fa_per_lipid, force_even=False)
        for real, imag in itertools.product(cc, db):
            fatty_acid = complex(real, imag)
            fa_profile[fatty_acid] += probability

    fa_set = set(fatty_acids)
    fa_profile = {lipid: prob for lipid, prob in fa_profile.items() if lipid in fa_set}

    total = sum(fa_profile.values())
    return {fatty_acid: prob / total for fatty_acid, prob in fa_profile.items()}



from typing import Any

from fatty_acylizer.util.optimization import Callback
from scipy.optimize import OptimizeResult
from sklearn.feature_extraction import DictVectorizer


def debug_serializer(
    result: OptimizeResult,
    history: list[Callback],
    settings: dict[str, Any],
    fa_vectorizer: DictVectorizer,
):
    print(result)
    print(history)
    print(settings)
    print(fa_vectorizer)



import itertools as it
import math
from collections import defaultdict

from fatty_acylizer.util.combinatorics import calculate_number_of_permutations
from fatty_acylizer.util.model import Profile
from fatty_acylizer.util.model import Result


class Model:
    def __init__(self, n_fa_per_lipid):
        self.n_fa_per_lipid = n_fa_per_lipid
        self.jac = False
        self.fatty_acids = None

    @property
    def settings(self):
        return {
            'jac': self.jac,
            'n_fa_per_lipid': self.n_fa_per_lipid,
            'fatty_acids': self.fatty_acids,
        }

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        params = ', '.join([f'{key}={value}' for key, value in self.settings.items()])
        return f'{cls_name}({params})'

    def __call__(self, fa_profile: Profile) -> Result:
        self.fatty_acids = list(fa_profile.keys())
        space = defaultdict(float)
        for combination in it.combinations_with_replacement(
            fa_profile, r=self.n_fa_per_lipid
        ):
            sum_feature = sum(combination)
            k_possible_permutations = calculate_number_of_permutations(combination)
            probability = (
                math.prod(fa_profile[item] for item in combination)
                * k_possible_permutations
            )
            space[sum_feature] += probability
        return Result(profile=dict(space), gradient=None)



import csv
import pathlib
from typing import Any
from typing import TextIO

import numpy as np
from fatty_acylizer.util.model import Model
from fatty_acylizer.util.optimization import Callback
from sklearn.feature_extraction import DictVectorizer


class FaProfileSerializer:
    def __init__(self, filename: str, vectorizer: DictVectorizer):
        self.filename = pathlib.Path(filename)
        self._vectorizer = vectorizer
        self.setup()

    def get_writer(self, file):
        writer = csv.DictWriter(
            file,
            fieldnames=self._vectorizer.feature_names_,
        )
        return writer

    def setup(self):
        with open(self.filename, 'w') as f:
            self.get_writer(f).writeheader()

    def add_callback(self, callback: Callback):
        with open(self.filename, 'a') as f:
            (profile,) = self._vectorizer.inverse_transform(callback.x.reshape(1, -1))
            self.get_writer(f).writerow(profile)

    def __call__(self, callback: Callback):
        return self.add_callback(callback)


class LipidProfileSerializer:
    def __init__(self, filename: str, model: Model):
        self.filename = pathlib.Path(filename)
        self.model = model
        self._fieldnames = list(sorted(str(lipid) for lipid in model.table[-1]))
        self.setup()

    def get_writer(self, file: TextIO):
        writer = csv.DictWriter(file, fieldnames=self._fieldnames)
        return writer

    def setup(self):
        with open(self.filename, 'w') as f:
            self.get_writer(f).writeheader()

    def add_callback(self, callback: Callback):  # noqa ARG002
        # call back is not used, but rather the last entry in the raw model table
        # This does not require the model to be reevaluated again
        lipid_profile = {
            str(lipid): prob for lipid, prob in self.model.table[-1].items()
        }
        with open(self.filename, 'a') as f:
            self.get_writer(f).writerow(lipid_profile)

    def __call__(self, callback: Callback):
        return self.add_callback(callback)


class DistanceSerializer:
    def __init__(self, filename: str):
        self.filename = pathlib.Path(filename)
        self.setup()

    def get_writer(self, file: TextIO):
        writer = csv.DictWriter(
            file,
            fieldnames=['distance'],
        )
        return writer

    def setup(self):
        with open(self.filename, 'w') as f:
            self.get_writer(f).writeheader()

    def add_callback(self, callback: Callback):
        with open(self.filename, 'a') as f:
            self.get_writer(f).writerow({'distance': callback.fun})

    def __call__(self, callback: Callback):
        return self.add_callback(callback)


class LogSerializer:
    def __init__(self, filename: str, settings: dict[str, Any]):
        self.filename = pathlib.Path(filename)
        self.setup(settings)
        self._iter = 0

    def get_writer(self, file: TextIO):
        return csv.DictWriter(file, fieldnames=['setting', 'value'], delimiter='\t')

    def setup(self, settings: dict[str, Any]):
        with open(self.filename, 'w') as f:
            f.write('Settings\n')
            f.write('--------\n\n')
            writer = self.get_writer(f)
            for setting, value in settings.items():
                if isinstance(value, np.ndarray):
                    value = ', '.join(str(v) for v in value)
                writer.writerow({'setting': setting, 'value': value})
            # for callbacks
            f.write('\nHistory\n')
            f.write('-------\n\n')

    def add_callback(self, callback: Callback):
        with open(self.filename, 'a') as f:
            f.write(f'Step {self._iter}:\n')
            self._iter += 1
            writer = self.get_writer(f)
            for setting, value in callback.report().items():
                if isinstance(value, np.ndarray):
                    value = ', '.join(str(v) for v in value)
                writer.writerow({'setting': setting, 'value': value})
            f.write('\n')

    def __call__(self, callback: Callback):
        return self.add_callback(callback)



import fatty_acylizer.util.serializer.csv as serializer
from fatty_acylizer.models.memo.model import Model
from fatty_acylizer.util.optimization import Optimization
from fatty_acylizer.util.suggestion import suggest_start_conditions

__all__ = [
    serializer,
    Model,
    Optimization,
    suggest_start_conditions
]



import typing
from collections import defaultdict

if typing.TYPE_CHECKING:
    from _typeshed import SupportsRichComparison  # noqa F401

from fatty_acylizer.util.feature import BasicNumber
from fatty_acylizer.util.model import Result


class Model:
    def __init__(self, fatty_acids: list[BasicNumber], n_fa_per_lipid: int):
        """fatty acids needs to be sorted."""
        self.jac = False
        self.n_fa_per_lipid = n_fa_per_lipid
        self.fatty_acids = fatty_acids
        self.table: list[dict] = []

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        params = ', '.join([f'{key}={value}' for key, value in self.settings.items()])
        return f'{cls_name}({params})'

    @property
    def settings(self):
        return {
            'jac': self.jac,
            'n_fa_per_lipid': self.n_fa_per_lipid,
            'fatty_acids': self.fatty_acids,
        }

    def __call__(self, fa_profile: dict[BasicNumber, float]) -> Result:
        if any(key not in self.fatty_acids for key in fa_profile):
            raise ValueError(fa_profile)
        self._initialize(fa_profile)
        for _ in range(self.n_fa_per_lipid - 1):
            self._propagate_to_next()
        return Result(profile=self.table[-1], gradient=None)

    def _initialize(self, fa_profile: dict[BasicNumber, float]) -> None:
        """Initialize self.table with zero containing arrays"""
        self.table = [fa_profile]

    def _propagate_to_next(self) -> None:
        n = len(self.table)
        new_row = defaultdict(lambda: 0)
        for def_val, def_prob in self.table[0].items():
            for pre_val, pre_prob in self.table[n - 1].items():
                new_row[pre_val + def_val] += pre_prob * def_prob
        self.table.append(dict(new_row))


if __name__ == '__main__':
    import pathlib

    with open(
        pathlib.Path('/home/janik/Documents/projects/predictor/data/fatty_acids.csv')
    ) as f:
        lines = (line.strip().split(',') for line in f)
        fatty_acids = [complex(real=int(real), imag=int(imag)) for real, imag in lines]

    model = Model(fatty_acids=fatty_acids, n_fa_per_lipid=3)



from dataclasses import dataclass
from dataclasses import field
from typing import Protocol
from typing import Self


class BasicNumber(Protocol):
    def __add__(self, other: Self, /) -> Self: ...

    def __radd__(self, other: Self, /) -> Self: ...

    def __sub__(self, other: Self, /) -> Self: ...


@dataclass(slots=True)
class Feature:
    carbon_chain_length: int
    double_bond_count: int
    labeled_carbons: int = field(default=0)

    def __add__(self, other: Self) -> Self:
        return self.__class__(
            carbon_chain_length=self.carbon_chain_length + other.carbon_chain_length,
            double_bond_count=self.double_bond_count + other.double_bond_count,
            labeled_carbons=self.labeled_carbons + other.labeled_carbons,
        )

    def __radd__(self, other: Self) -> Self:
        if other == 0:
            # for use with sum(...)
            return self
        return self + other

    def __neg__(self) -> Self:
        return self.__class__(
            carbon_chain_length=-self.carbon_chain_length,
            double_bond_count=-self.double_bond_count,
            labeled_carbons=-self.labeled_carbons,
        )

    def __sub__(self, other: Self) -> Self:
        return self.__class__(
            carbon_chain_length=self.carbon_chain_length - other.carbon_chain_length,
            double_bond_count=self.double_bond_count - other.double_bond_count,
            labeled_carbons=self.labeled_carbons - other.labeled_carbons,
        )


if __name__ == '__main__':
    number: BasicNumber = 1 + 2j
    number2: BasicNumber = Feature(1, 2, 1)
    del number
    del number2



import functools as fn
import math
from collections import Counter
from collections.abc import Iterable


def calculate_number_of_permutations(collection: Iterable) -> int:
    count = Counter(collection)
    n_permutations = factorial(sum(count.values())) / math.prod(
        factorial(rep) for rep in count.values()
    )
    if n_permutations.is_integer():
        return int(n_permutations)
    raise ValueError(f'{n_permutations} is not an integer')


factorial = fn.lru_cache(math.factorial)



from typing import Any
from typing import NamedTuple
from typing import Protocol

from fatty_acylizer.util.feature import BasicNumber


Profile = dict[BasicNumber, float]


class Result(NamedTuple):
    profile: Profile
    gradient: list[Profile] | None


class Model(Protocol):
    jac: bool
    n_fa_per_lipid: int
    fatty_acids: list[BasicNumber]
    settings: dict[str, Any]

    def __call__(self, fa_profile: Profile) -> Result: ...
